import { Level } from 'level';
const levelDbOptions = { valueEncoding: 'json' };
export function createDatabase(directoryPath) {
    const db = new Level(directoryPath);
    async function deleteIncompleteJobsAsync() {
        const deleteOperations = [];
        for await (const [key, result] of db.iterator(levelDbOptions)) {
            if (result.status !== 'COMPLETE') {
                deleteOperations.push({ key, type: 'del' });
            }
        }
        db.batch(deleteOperations);
    }
    async function getJobStatusAsync(type, id) {
        try {
            const key = createKey(type, id);
            const result = await db.get(key, levelDbOptions);
            return result;
        }
        catch (error) {
            return null;
        }
    }
    async function setJobStatusToQueuedAsync(type, id) {
        const result = await getJobStatusAsync(type, id);
        if (result !== null) {
            throw new Error('`result` must be `null`');
        }
        const key = createKey(type, id);
        const value = {
            resultUrl: `${type}/${id}`,
            status: 'QUEUED'
        };
        return db.put(key, value, levelDbOptions);
    }
    async function setJobStatusToInProgressAsync(type, id, progress) {
        const result = await getJobStatusAsync(type, id);
        if (result === null) {
            throw new Error('`result` is `null`');
        }
        if (result.status === 'COMPLETE') {
            throw new Error("`result` cannot be 'COMPLETE'");
        }
        const imageUrls = [];
        const completedImages = progress.currentImageProgress === 1
            ? progress.currentImageIndex
            : Math.max(0, progress.currentImageIndex - 1);
        let i = 0;
        while (i < completedImages) {
            imageUrls.push(`${type}/${id}/${i + 1}.png`);
            i += 1;
        }
        const key = createKey(type, id);
        const value = {
            imageUrls,
            progress,
            resultUrl: result.resultUrl,
            status: 'IN_PROGRESS'
        };
        return db.put(key, value, levelDbOptions);
    }
    async function setJobStatusToDoneAsync(type, id) {
        const result = await getJobStatusAsync(type, id);
        if (result === null) {
            throw new Error('`result` is `null`');
        }
        if (result.status !== 'IN_PROGRESS') {
            throw new Error("`result` must be 'IN_PROGRESS'");
        }
        const key = createKey(type, id);
        const value = {
            imageUrls: result.imageUrls,
            resultUrl: result.resultUrl,
            status: 'COMPLETE'
        };
        return db.put(key, value, levelDbOptions);
    }
    return {
        deleteIncompleteJobsAsync,
        getJobStatusAsync,
        setJobStatusToDoneAsync,
        setJobStatusToInProgressAsync,
        setJobStatusToQueuedAsync
    };
}
function createKey(type, id) {
    return `${type}/${id}`;
}
//# sourceMappingURL=create-database.js.map